/*****************************************************************************
 * grab-dvb.c: support for Digital Video Broadcast
 *****************************************************************************
 * $Id: grab-dvb.c,v 1.48 2008/09/12 21:34:53 pingus77 Exp $
 *****************************************************************************
 * Program under GNU General Public License (see ../COPYING)
 *****************************************************************************/

#include "config.h"

#ifdef HAVE_DVB

#include "config.h"
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/times.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <math.h>
#include <ctype.h>

#include <linux/dvb/frontend.h>
#include <linux/dvb/dmx.h>

#if HAVE_LIBAVCODEC_AVCODEC_H
# include <libavcodec/avcodec.h>
# include <libavcodec/audioconvert.h>
#else
# include "avcodec.h"
# include "audioconvert.h"
#endif
#if HAVE_LIBAVFORMAT_AVFORMAT_H
# include <libavformat/avformat.h>
#else
# include "avformat.h"
#endif
#if HAVE_LIBAVUTIL_COMMON_H
# include <libavutil/common.h>
#else
# include "common.h"
#endif
#ifdef HAVE_SWSCALE
#if HAVE_LIBSWSCALE_SWSCALE_H
# include <libswscale/swscale.h>
#else
# include "swscale.h"
#endif
#endif

#include "divx.h"
#include "strtab.h"
#include "grab.h"
#include "colorspace.h"
#include "audio.h"
#include "mixer.h"
#include "memcpy.h"
#include "channel.h"
#include "dvbtools.h"
#include "avilib.h"

extern int devicemanager_tadapter ;
extern int filenoaudioout;
extern int use_file_audio;
extern int debug;
extern avi_t *   videofile ;
extern double audiostamp;
extern double delay; // delay between two images
extern clock_t t_start;
extern struct tms start;
extern pthread_t audio_thread;
extern pthread_mutex_t file_mutex;
extern char* audiobuf;
extern int audiobuflen;
extern int audio_buf_max_len;
extern int use_file_audio;
extern int filenoaudioout;
extern void video_set_capture_size(int w, int h);
extern int nbufs;
extern AVInputFormat mpegts_demuxer;
extern AVCodec mpeg2video_decoder;
extern AVCodec mp2_decoder;
extern char* avifilename;

int frameFinished;
int  dvb_grabber_running=0;
static int stream_opened=0;

/**** Video **/
static    AVFrame         *pFrameRGB;
static    int             numBytesRGB;
static    uint8_t         *bufferRGB;
static    AVFrame         *pFrameSRC;
static    int             numBytesSRC;
static    uint8_t         *bufferSRC;

/**** Audio **/
int 	  out_size_audio;

static    AVFrame         *pFrameRGB_resized=NULL;
#ifndef HAVE_SWSCALE
static    int             numBytesRGB_resized=0;
#endif
static    int             width_resized=0;
static    int             height_resized=0;
static    uint8_t         *bufferRGB_resized=NULL;
static    AVFrame         *pFrameYUV_resized=NULL;
#ifndef HAVE_SWSCALE
static    int             numBytesYUV_resized=0;
#endif
static    uint8_t         *bufferYUV_resized=NULL;

#ifndef HAVE_SWSCALE

/*************************************************************************/
/*To avoid any warning as FFMPEG deprecated then removed theses functions*/

typedef struct ImgReSampleContext ImgReSampleContext;

ImgReSampleContext *img_resample_init(int output_width, int output_height, int input_width, int input_height);
void img_resample(ImgReSampleContext *s, AVPicture *output, const AVPicture *input);
int img_convert(AVPicture *dst, int dst_pix_fmt, AVPicture *src, int pix_fmt, int width, int height);

/***********************************************************************/

static struct ImgReSampleContext* imgResCtx=NULL;

#else

static struct SwsContext* swscale_context=NULL;

#endif

static pthread_t event_loop_thread;
static int  event_loop_thread_started=0;
static pthread_mutex_t stream_open_mutex;
static pthread_mutex_t new_pict_mutex;
static pthread_mutex_t audio_buf_mutex;
static int new_picture_ready;
static int new_audiobuf_ready;

/* dvb parameters */
static struct dvb_frontend_parameters frontend_param;
static int vpid, apid, spid ;
static int dvr = 0;
static int frontend_fd, audio_fd, video_fd, vdemuxfd, ademuxfd,sdemuxfd;
int subtit_fd; // used by AleVT

/**************** BEGIN FFPLAY SDL PLAYER ******************************/

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#include <X11/Xlib.h>

#define MAX_VIDEOQ_SIZE (5 * 256 * 1024)
#define MAX_AUDIOQ_SIZE (5 * 16 * 1024)
#define MAX_SUBTITLEQ_SIZE sizeof(wchar_t)*(5 * 16 * 1024)

/* SDL audio buffer size, in samples. Should be small to have precise
   A/V sync as SDL does not have hardware buffer fullness info. */
#define SDL_AUDIO_BUFFER_SIZE 1024

/* no AV sync correction is done if below the AV sync threshold */
#define AV_SYNC_THRESHOLD 0.01
/* no AV correction is done if too big error */
#define AV_NOSYNC_THRESHOLD 10.0

/* maximum audio speed change to get correct sync */
#define SAMPLE_CORRECTION_PERCENT_MAX 10

/* we use about AUDIO_DIFF_AVG_NB A-V differences to make the average */
#define AUDIO_DIFF_AVG_NB   20

/* NOTE: the size must be big enough to compensate the hardware audio buffersize size */
#define SAMPLE_ARRAY_SIZE (2*65536)

typedef struct PacketQueue {
    AVPacketList *first_pkt, *last_pkt;
    int nb_packets;
    int size;
    int abort_request;
    SDL_mutex *mutex;
    SDL_cond *cond;
} PacketQueue;

#define VIDEO_PICTURE_QUEUE_SIZE 1
#define SUBPICTURE_QUEUE_SIZE 4

// SYNCH Rev. 1137
typedef struct VideoPicture {
    double pts;                                  ///<presentation time stamp for this picture
    //SDL_Overlay *bmp; // XDTV
    int width, height; /* source height & width */
    int allocated;
} VideoPicture;

// SYNCH Rev. 1137
typedef struct SubPicture {
    double pts; /* presentation time stamp for this picture */
    AVSubtitle sub;
} SubPicture;

// SYNCH Rev. 1137
enum {
    AV_SYNC_AUDIO_MASTER, /* default choice */
    AV_SYNC_VIDEO_MASTER,
    AV_SYNC_EXTERNAL_CLOCK, /* synchronize to an external clock */
};

// SYNCH Rev. 1137
typedef struct VideoState {
    SDL_Thread *parse_tid;
    SDL_Thread *video_tid;
    AVInputFormat *iformat;
    int no_background;
    int abort_request;
    int paused;
    int last_paused;
    int seek_req;
    int seek_flags;
    int64_t seek_pos;
    AVFormatContext *ic;
    int dtg_active_format;

    int audio_stream;

    int av_sync_type;
    double external_clock; /* external clock base */
    int64_t external_clock_time;

    double audio_clock;
    double audio_diff_cum; /* used for AV difference average computation */
    double audio_diff_avg_coef;
    double audio_diff_threshold;
    int audio_diff_avg_count;
    AVStream *audio_st;
    PacketQueue audioq;
    int audio_hw_buf_size;
    /* samples output by the codec. we reserve more space for avsync
       compensation */
    DECLARE_ALIGNED(16,uint8_t,audio_buf1[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2]);
    DECLARE_ALIGNED(16,uint8_t,audio_buf2[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2]);
    uint8_t *audio_buf;
    unsigned int audio_buf_size; /* in bytes */
    int audio_buf_index; /* in bytes */
    AVPacket audio_pkt;
    uint8_t *audio_pkt_data;
    int audio_pkt_size;
    enum SampleFormat audio_src_fmt;
    AVAudioConvert *reformat_ctx;

    int show_audio; /* if true, display audio samples */
    int16_t sample_array[SAMPLE_ARRAY_SIZE];
    int sample_array_index;
    int last_i_start;

    SDL_Thread *subtitle_tid;
    int subtitle_stream;
    int subtitle_stream_changed;
    AVStream *subtitle_st;
    PacketQueue subtitleq;
    SubPicture subpq[SUBPICTURE_QUEUE_SIZE];
    int subpq_size, subpq_rindex, subpq_windex;
    SDL_mutex *subpq_mutex;
    SDL_cond *subpq_cond;

    double frame_timer;
    double frame_last_pts;
    double frame_last_delay;
    double video_clock;                          ///<pts of last decoded frame / predicted pts of next decoded frame
    int video_stream;
    AVStream *video_st;
    PacketQueue videoq;
    double video_current_pts;                    ///<current displayed pts (different from video_clock if frame fifos are used)
    int64_t video_current_pts_time;              ///<time (av_gettime) at which we updated video_current_pts - used to have running video pts
    VideoPicture pictq[VIDEO_PICTURE_QUEUE_SIZE];
    int pictq_size, pictq_rindex, pictq_windex;
    SDL_mutex *pictq_mutex;
    SDL_cond *pictq_cond;

    //    QETimer *video_timer;
    char filename[1024];
    int width, height, xleft, ytop;
} VideoState;

static int audio_write_get_buf_size(VideoState *is);

/* options specified by the user */
static const char *input_filename;

static int audio_disable;
static int video_disable;
static int wanted_audio_stream= 0;
static int wanted_video_stream= 0;
static int seek_by_bytes;
static int display_disable;
static int av_sync_type = AV_SYNC_AUDIO_MASTER;
static int64_t start_time = AV_NOPTS_VALUE;
static int debug_mv = 0;
static int step = 0;
static int thread_count = 1;
static int workaround_bugs = 1;
static int fast = 0;
static int genpts = 0;
static int lowres = 0;
static int idct = FF_IDCT_AUTO;
static enum AVDiscard skip_frame= AVDISCARD_DEFAULT;
static enum AVDiscard skip_idct= AVDISCARD_DEFAULT;
static enum AVDiscard skip_loop_filter= AVDISCARD_DEFAULT;
static int error_recognition = FF_ER_CAREFUL;
static int error_concealment = 3;

/* current context */
static VideoState *cur_stream=NULL;
static int64_t audio_callback_time;

AVPacket flush_pkt;

#define FF_ALLOC_EVENT   (SDL_USEREVENT)
#define FF_REFRESH_EVENT (SDL_USEREVENT + 1)
#define FF_QUIT_EVENT    (SDL_USEREVENT + 2)

// SYNCH Rev. 1137
/* packet queue handling */
static void packet_queue_init(PacketQueue *q)
{
    memset(q, 0, sizeof(PacketQueue));
    q->mutex = SDL_CreateMutex();
    q->cond = SDL_CreateCond();
}

// SYNCH Rev. 1137
static void packet_queue_flush(PacketQueue *q)
{
    AVPacketList *pkt, *pkt1;

    SDL_LockMutex(q->mutex);
    for(pkt = q->first_pkt; pkt != NULL; pkt = pkt1) {
        pkt1 = pkt->next;
        av_free_packet(&pkt->pkt);
        av_freep(&pkt);
    }
    q->last_pkt = NULL;
    q->first_pkt = NULL;
    q->nb_packets = 0;
    q->size = 0;
    SDL_UnlockMutex(q->mutex);
}


// SYNCH Rev. 1137
static void packet_queue_end(PacketQueue *q)
{
    packet_queue_flush(q);
    SDL_DestroyMutex(q->mutex);
    SDL_DestroyCond(q->cond);
}

// SYNCH Rev. 1137
static int packet_queue_put(PacketQueue *q, AVPacket *pkt)
{
    AVPacketList *pkt1;

    /* duplicate the packet */
    if (pkt!=&flush_pkt && av_dup_packet(pkt) < 0)
        return -1;

    pkt1 = av_malloc(sizeof(AVPacketList));
    if (!pkt1)
        return -1;
    pkt1->pkt = *pkt;
    pkt1->next = NULL;


    SDL_LockMutex(q->mutex);

    if (!q->last_pkt)

        q->first_pkt = pkt1;
    else
        q->last_pkt->next = pkt1;
    q->last_pkt = pkt1;
    q->nb_packets++;
    q->size += pkt1->pkt.size;
    /* XXX: should duplicate packet data in DV case */
    SDL_CondSignal(q->cond);

    SDL_UnlockMutex(q->mutex);
    return 0;
}

// SYNCH Rev. 1137
static void packet_queue_abort(PacketQueue *q)
{
    SDL_LockMutex(q->mutex);

    q->abort_request = 1;

    SDL_CondSignal(q->cond);

    SDL_UnlockMutex(q->mutex);
}

// SYNCH Rev. 1137
/* return < 0 if aborted, 0 if no packet and > 0 if packet.  */
static int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block)
{
    AVPacketList *pkt1;
    int ret;

    SDL_LockMutex(q->mutex);

    for(;;) {
        if (q->abort_request) {
            ret = -1;
            break;
        }

        pkt1 = q->first_pkt;
        if (pkt1) {
            q->first_pkt = pkt1->next;
            if (!q->first_pkt)
                q->last_pkt = NULL;
            q->nb_packets--;
            q->size -= pkt1->pkt.size;
            *pkt = pkt1->pkt;
            av_free(pkt1);
            ret = 1;
            break;
        } else if (!block) {
            ret = 0;
            break;
        } else {
            SDL_CondWait(q->cond, q->mutex);
        }
    }
    SDL_UnlockMutex(q->mutex);
    return ret;
}

// SYNCH Rev. 1137
static void free_subpicture(SubPicture *sp)
{
    int i;

    for (i = 0; i < sp->sub.num_rects; i++)
    {
        av_free(sp->sub.rects[i].bitmap);
        av_free(sp->sub.rects[i].rgba_palette);
    }

    av_free(sp->sub.rects);

    memset(&sp->sub, 0, sizeof(AVSubtitle));
}

// SYNCH Rev. 1137
static Uint32 sdl_refresh_timer_cb(Uint32 interval, void *opaque)
{
    SDL_Event event;
    event.type = FF_REFRESH_EVENT;
    event.user.data1 = opaque;
    SDL_PushEvent(&event);
    return 0; /* 0 means stop timer */
}

// SYNCH Rev. 1137
/* schedule a video refresh in 'delay' ms */
static void schedule_refresh(VideoState *is, int delay)
{
    if(!delay) delay=1; //SDL seems to be buggy when the delay is 0
    SDL_AddTimer(delay, sdl_refresh_timer_cb, is);
}

// SYNCH Rev. 1137
/* get the current audio clock value */
static double get_audio_clock(VideoState *is)
{
    double pts;
    int hw_buf_size, bytes_per_sec;
    pts = is->audio_clock;
    hw_buf_size = audio_write_get_buf_size(is);
    bytes_per_sec = 0;
    if (is->audio_st) {
        bytes_per_sec = is->audio_st->codec->sample_rate *
            2 * is->audio_st->codec->channels;
    }
    if (bytes_per_sec)
        pts -= (double)hw_buf_size / bytes_per_sec;
    return pts;
}

// SYNCH Rev. 1137
/* get the current video clock value */
static double get_video_clock(VideoState *is)
{
    double delta;
    if (is->paused) {
        delta = 0;
    } else {
        delta = (av_gettime() - is->video_current_pts_time) / 1000000.0;
    }
    return is->video_current_pts + delta;
}

// SYNCH Rev. 1137
/* get the current external clock value */
static double get_external_clock(VideoState *is)
{
    int64_t ti;
    ti = av_gettime();
    return is->external_clock + ((ti - is->external_clock_time) * 1e-6);
}

// SYNCH Rev. 1137
/* get the current master clock value */
static double get_master_clock(VideoState *is)
{
    double val;

    if (is->av_sync_type == AV_SYNC_VIDEO_MASTER) {
        if (is->video_st)
            val = get_video_clock(is);
        else
            val = get_audio_clock(is);
    } else if (is->av_sync_type == AV_SYNC_AUDIO_MASTER) {
        if (is->audio_st)
            val = get_audio_clock(is);
        else
            val = get_video_clock(is);
    } else {
        val = get_external_clock(is);
    }
    return val;
}

// SYNCH Rev. 1137
/* seek in the stream */
static void stream_seek(VideoState *is, int64_t pos, int rel)
{
    if (!is->seek_req) {
        is->seek_pos = pos;
        is->seek_flags = rel < 0 ? AVSEEK_FLAG_BACKWARD : 0;
        if (seek_by_bytes)
            is->seek_flags |= AVSEEK_FLAG_BYTE;
        is->seek_req = 1;
    }
}

// SYNCH Rev. 1137
/* pause or resume the video */
static void stream_pause(VideoState *is)
{
    is->paused = !is->paused;
    if (!is->paused) {
        is->video_current_pts = get_video_clock(is);
        is->frame_timer += (av_gettime() - is->video_current_pts_time) / 1000000.0;
    }
}

// SYNCH Rev. 1137
/* called to display each frame */
static void video_refresh_timer(void *opaque)
{
    VideoState *is = opaque;
    VideoPicture *vp;
    double actual_delay, delay, sync_threshold, ref_clock, diff;

    SubPicture *sp, *sp2;

    if (is->video_st) {
        if (is->pictq_size == 0) {
            /* if no picture, need to wait */
            schedule_refresh(is, 1);
        } else {
            /* dequeue the picture */
            vp = &is->pictq[is->pictq_rindex];

            /* update current video pts */
            is->video_current_pts = vp->pts;
            is->video_current_pts_time = av_gettime();

            /* compute nominal delay */
            delay = vp->pts - is->frame_last_pts;
            if (delay <= 0 || delay >= 2.0) {
                /* if incorrect delay, use previous one */
                delay = is->frame_last_delay;
            }
            is->frame_last_delay = delay;
            is->frame_last_pts = vp->pts;

            /* update delay to follow master synchronisation source */
            if (((is->av_sync_type == AV_SYNC_AUDIO_MASTER && is->audio_st) ||
                 is->av_sync_type == AV_SYNC_EXTERNAL_CLOCK)) {
                /* if video is slave, we try to correct big delays by
                   duplicating or deleting a frame */
                ref_clock = get_master_clock(is);
                diff = vp->pts - ref_clock;

                /* skip or repeat frame. We take into account the
                   delay to compute the threshold. I still don't know
                   if it is the best guess */
                sync_threshold = FFMAX(AV_SYNC_THRESHOLD, delay);
                if (fabs(diff) < AV_NOSYNC_THRESHOLD) {
                    if (diff <= -sync_threshold)
                        delay = 0;
                    else if (diff >= sync_threshold)
                        delay = 2 * delay;
                }
            }

            is->frame_timer += delay;
            /* compute the REAL delay (we need to do that to avoid
               long term errors */
            actual_delay = is->frame_timer - (av_gettime() / 1000000.0);
            if (actual_delay < 0.010) {
                /* XXX: should skip picture */
                actual_delay = 0.010;
            }
            /* launch timer for next picture */
            schedule_refresh(is, (int)(actual_delay * 1000 + 0.5));

#if defined(DEBUG_SYNC)
            printf("video: delay=%0.3f actual_delay=%0.3f pts=%0.3f A-V=%f\n",
                    delay, actual_delay, vp->pts, -diff);
#endif

            if(is->subtitle_st) {
                if (is->subtitle_stream_changed) {
                    SDL_LockMutex(is->subpq_mutex);

                    while (is->subpq_size) {
                        free_subpicture(&is->subpq[is->subpq_rindex]);

                        /* update queue size and signal for next picture */
                        if (++is->subpq_rindex == SUBPICTURE_QUEUE_SIZE)
                            is->subpq_rindex = 0;

                        is->subpq_size--;
                    }
                    is->subtitle_stream_changed = 0;

                    SDL_CondSignal(is->subpq_cond);
                    SDL_UnlockMutex(is->subpq_mutex);
                } else {
                    if (is->subpq_size > 0) {
                        sp = &is->subpq[is->subpq_rindex];

                        if (is->subpq_size > 1)
                            sp2 = &is->subpq[(is->subpq_rindex + 1) % SUBPICTURE_QUEUE_SIZE];
                        else
                            sp2 = NULL;

                        if ((is->video_current_pts > (sp->pts + ((float) sp->sub.end_display_time / 1000)))
                                || (sp2 && is->video_current_pts > (sp2->pts + ((float) sp2->sub.start_display_time / 1000))))
                        {
                            free_subpicture(sp);

                            /* update queue size and signal for next picture */
                            if (++is->subpq_rindex == SUBPICTURE_QUEUE_SIZE)
                                is->subpq_rindex = 0;

                            SDL_LockMutex(is->subpq_mutex);
                            is->subpq_size--;
                            SDL_CondSignal(is->subpq_cond);
                            SDL_UnlockMutex(is->subpq_mutex);
                        }
                    }
                }
            }

            /* update queue size and signal for next picture */
            if (++is->pictq_rindex == VIDEO_PICTURE_QUEUE_SIZE)
                is->pictq_rindex = 0;

            SDL_LockMutex(is->pictq_mutex);
            is->pictq_size--;
            SDL_CondSignal(is->pictq_cond);
            SDL_UnlockMutex(is->pictq_mutex);
        }
    } else if (is->audio_st) {
        /* draw the next audio frame */

        schedule_refresh(is, 40);

        /* if only audio stream, then display the audio bars (better
           than nothing, just to test the implementation */

        /* display picture */
        //video_display(is); //XDTV
    } else {
        schedule_refresh(is, 100);
    }
//    if (1/*show_status*/) {
    if (0) {
        static int64_t last_time;
        int64_t cur_time;
        int aqsize, vqsize, sqsize;
        double av_diff;

        cur_time = av_gettime();
        if (!last_time || (cur_time - last_time) >= 500 * 1000) {
            aqsize = 0;
            vqsize = 0;
            sqsize = 0;
            if (is->audio_st)
                aqsize = is->audioq.size;
            if (is->video_st)
                vqsize = is->videoq.size;
            if (is->subtitle_st)
                sqsize = is->subtitleq.size;
            av_diff = 0;
            if (is->audio_st && is->video_st)
                av_diff = get_audio_clock(is) - get_video_clock(is);
            printf("%7.2f A-V:%7.3f aq=%5dKB vq=%5dKB sq=%5dB    \r",
                   get_master_clock(is), av_diff, aqsize / 1024, vqsize / 1024, sqsize);
            fflush(stdout);
            last_time = cur_time;
        }
    }
}


// SYNCH Rev. 1137
/* allocate a picture (needs to do that in main thread to avoid
   potential locking problems */
static void alloc_picture(void *opaque)
{
    VideoState *is = opaque;
    VideoPicture *vp;

    vp = &is->pictq[is->pictq_windex];

// XDTV (no bmp)
/*
    if (vp->bmp)
        SDL_FreeYUVOverlay(vp->bmp);

    vp->bmp = SDL_CreateYUVOverlay(is->video_st->codec->width,
                                   is->video_st->codec->height,
                                   SDL_YV12_OVERLAY,
                                   screen);
*/

    vp->width = is->video_st->codec->width;
    vp->height = is->video_st->codec->height;

    SDL_LockMutex(is->pictq_mutex);
    vp->allocated = 1;
    SDL_CondSignal(is->pictq_cond);
    SDL_UnlockMutex(is->pictq_mutex);
}


// SYNCH Rev. 1137
/**
 *
 * @param pts the dts of the pkt / pts of the frame and guessed if not known
 */
static int queue_picture(VideoState *is, AVFrame *src_frame, double pts)
{
    VideoPicture *vp;
    //int dst_pix_fmt; //XDTV
    //AVPicture pict; //XDTV
    //static struct SwsContext *img_convert_ctx; //XDTV

    /* wait until we have space to put a new picture */
    SDL_LockMutex(is->pictq_mutex);
    while (is->pictq_size >= VIDEO_PICTURE_QUEUE_SIZE &&
           !is->videoq.abort_request) {
        SDL_CondWait(is->pictq_cond, is->pictq_mutex);
    }
    SDL_UnlockMutex(is->pictq_mutex);

    if (is->videoq.abort_request)
        return -1;

    vp = &is->pictq[is->pictq_windex];

    /* alloc or resize hardware picture buffer */
    if (/*!vp->bmp || // XDTV (no bmp) */
        vp->width != is->video_st->codec->width ||
        vp->height != is->video_st->codec->height) {
        SDL_Event event;

        vp->allocated = 0;

        /* the allocation must be done in the main thread to avoid
           locking problems */
        event.type = FF_ALLOC_EVENT;
        event.user.data1 = is;
        SDL_PushEvent(&event);

        /* wait until the picture is allocated */
        SDL_LockMutex(is->pictq_mutex);
        while (!vp->allocated && !is->videoq.abort_request) {
            SDL_CondWait(is->pictq_cond, is->pictq_mutex);
        }
        SDL_UnlockMutex(is->pictq_mutex);

        if (is->videoq.abort_request)
            return -1;
    }

/* a picture is ready to be displayed */

// XDTV --------------------------------------------------------------------------------------------
        pthread_mutex_lock(&new_pict_mutex);

        av_picture_copy((AVPicture *)pFrameSRC ,(AVPicture*) (src_frame), is->video_st->codec->pix_fmt ,
                is->video_st->codec->width, is->video_st->codec->height);

        new_picture_ready=1;
        pthread_mutex_unlock(&new_pict_mutex);
// XDTV ----------------------------------------------------------------------------------------END
        vp->pts = pts;

        /* now we can update the picture count */
        if (++is->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE)
            is->pictq_windex = 0;
        SDL_LockMutex(is->pictq_mutex);
        is->pictq_size++;
        SDL_UnlockMutex(is->pictq_mutex);

// XDTV (no bmp so this part is skipped!)
    /* if the frame is not skipped, then display it */
/*
    if (vp->bmp) {
        // get a pointer on the bitmap
        SDL_LockYUVOverlay (vp->bmp);

        dst_pix_fmt = PIX_FMT_YUV420P;
        pict.data[0] = vp->bmp->pixels[0];
        pict.data[1] = vp->bmp->pixels[2];
        pict.data[2] = vp->bmp->pixels[1];

        pict.linesize[0] = vp->bmp->pitches[0];
        pict.linesize[1] = vp->bmp->pitches[2];
        pict.linesize[2] = vp->bmp->pitches[1];
        img_convert_ctx = sws_getCachedContext(img_convert_ctx,
            is->video_st->codec->width, is->video_st->codec->height,
            is->video_st->codec->pix_fmt,
            is->video_st->codec->width, is->video_st->codec->height,
            dst_pix_fmt, sws_flags, NULL, NULL, NULL);
        if (img_convert_ctx == NULL) {
            fprintf(stderr, "Cannot initialize the conversion context\n");
            exit(1);
        }
        sws_scale(img_convert_ctx, src_frame->data, src_frame->linesize,
                  0, is->video_st->codec->height, pict.data, pict.linesize);
        // update the bitmap content
        SDL_UnlockYUVOverlay(vp->bmp);

        vp->pts = pts;

        // now we can update the picture count
        if (++is->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE)
            is->pictq_windex = 0;
        SDL_LockMutex(is->pictq_mutex);
        is->pictq_size++;
        SDL_UnlockMutex(is->pictq_mutex);
    }
*/
    return 0;
}

// SYNCH Rev. 1137
/**
 * compute the exact PTS for the picture if it is omitted in the stream
 * @param pts1 the dts of the pkt / pts of the frame
 */
static int output_picture2(VideoState *is, AVFrame *src_frame, double pts1)
{
    double frame_delay, pts;
    
    pts = pts1;

    if (pts != 0) {
        /* update video clock with pts, if present */
        is->video_clock = pts;
    } else {
        pts = is->video_clock;
    }
    /* update video clock for next frame */
    frame_delay = av_q2d(is->video_st->codec->time_base);
    /* for MPEG2, the frame can be repeated, so we update the
       clock accordingly */
    frame_delay += src_frame->repeat_pict * (frame_delay * 0.5);
    is->video_clock += frame_delay;

#if defined(DEBUG_SYNC) && 0
    {
        int ftype;
        if (src_frame->pict_type == FF_B_TYPE)
            ftype = 'B';
        else if (src_frame->pict_type == FF_I_TYPE)
            ftype = 'I';
        else
            ftype = 'P';
        printf("frame_type=%c clock=%0.3f pts=%0.3f\n",
               ftype, pts, pts1);
    }
#endif
    return queue_picture(is, src_frame, pts);
}

// SYNCH Rev. 1137
static int video_thread(void *arg)
{
    VideoState *is = arg;
    AVPacket pkt1, *pkt = &pkt1;
    int len1, got_picture;
    AVFrame *frame= avcodec_alloc_frame();
    double pts;

    for(;;) {
        while (is->paused && !is->videoq.abort_request) {
            SDL_Delay(10);
        }
        if (packet_queue_get(&is->videoq, pkt, 1) < 0)
            break;

        if(pkt->data == flush_pkt.data){
            avcodec_flush_buffers(is->video_st->codec);
            continue;
        }

        /* NOTE: ipts is the PTS of the _first_ picture beginning in
           this packet, if any */
        is->video_st->codec->reordered_opaque= pkt->pts;
        len1 = avcodec_decode_video(is->video_st->codec,
                                    frame, &got_picture,
                                    pkt->data, pkt->size);

        if(   (/*decoder_reorder_pts || // XDTV dfunction doesn't exist*/ pkt->dts == AV_NOPTS_VALUE) 
           && frame->reordered_opaque != AV_NOPTS_VALUE)
            pts= frame->reordered_opaque;
        else if(pkt->dts != AV_NOPTS_VALUE)
            pts= pkt->dts;
        else
            pts= 0;
        pts *= av_q2d(is->video_st->time_base);

//            if (len1 < 0)
//                break;
        if (got_picture) {
            if (output_picture2(is, frame, pts) < 0)
                goto the_end;
        }
        av_free_packet(pkt);
        if (step)
            if (cur_stream)
                stream_pause(cur_stream);
    }
 the_end:
    av_free(frame);
    return 0;
}

// SYNCH Rev. 1137
/* return the new audio buffer size (samples can be added or deleted
   to get better sync if video or external master clock) */
static int synchronize_audio(VideoState *is, short *samples,
                             int samples_size1, double pts)
{
    int n, samples_size;
    double ref_clock;

    n = 2 * is->audio_st->codec->channels;
    samples_size = samples_size1;

    /* if not master, then we try to remove or add samples to correct the clock */
    if (((is->av_sync_type == AV_SYNC_VIDEO_MASTER && is->video_st) ||
         is->av_sync_type == AV_SYNC_EXTERNAL_CLOCK)) {
        double diff, avg_diff;
        int wanted_size, min_size, max_size, nb_samples;

        ref_clock = get_master_clock(is);
        diff = get_audio_clock(is) - ref_clock;

        if (diff < AV_NOSYNC_THRESHOLD) {
            is->audio_diff_cum = diff + is->audio_diff_avg_coef * is->audio_diff_cum;
            if (is->audio_diff_avg_count < AUDIO_DIFF_AVG_NB) {
                /* not enough measures to have a correct estimate */
                is->audio_diff_avg_count++;
            } else {
                /* estimate the A-V difference */
                avg_diff = is->audio_diff_cum * (1.0 - is->audio_diff_avg_coef);

                if (fabs(avg_diff) >= is->audio_diff_threshold) {
                    wanted_size = samples_size + ((int)(diff * is->audio_st->codec->sample_rate) * n);
                    nb_samples = samples_size / n;

                    min_size = ((nb_samples * (100 - SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
                    max_size = ((nb_samples * (100 + SAMPLE_CORRECTION_PERCENT_MAX)) / 100) * n;
                    if (wanted_size < min_size)
                        wanted_size = min_size;
                    else if (wanted_size > max_size)
                        wanted_size = max_size;

                    /* add or remove samples to correction the synchro */
                    if (wanted_size < samples_size) {
                        /* remove samples */
                        samples_size = wanted_size;
                    } else if (wanted_size > samples_size) {
                        uint8_t *samples_end, *q;
                        int nb;

                        /* add samples */
                        nb = (samples_size - wanted_size);
                        samples_end = (uint8_t *)samples + samples_size - n;
                        q = samples_end + n;
                        while (nb > 0) {
                            memcpy(q, samples_end, n);
                            q += n;
                            nb -= n;
                        }
                        samples_size = wanted_size;
                    }
                }
#if 0
                printf("diff=%f adiff=%f sample_diff=%d apts=%0.3f vpts=%0.3f %f\n",
                       diff, avg_diff, samples_size - samples_size1,
                       is->audio_clock, is->video_clock, is->audio_diff_threshold);
#endif
            }
        } else {
            /* too big difference : may be initial PTS errors, so
               reset A-V filter */
            is->audio_diff_avg_count = 0;
            is->audio_diff_cum = 0;
        }
    }

    return samples_size;
}

// SYNCH Rev. 14508
/* decode one audio frame and returns its uncompressed size */
static int audio_decode_frame(VideoState *is, double *pts_ptr)
{
    AVPacket *pkt = &is->audio_pkt;
    AVCodecContext *dec= is->audio_st->codec;
    int n, len1, data_size;
    double pts;

    for(;;) {
        /* NOTE: the audio packet can contain several frames */
        while (is->audio_pkt_size > 0) {
            data_size = sizeof(is->audio_buf1);
            len1 = avcodec_decode_audio2(dec,
                                        (int16_t *)is->audio_buf1, &data_size,
                                         is->audio_pkt_data, is->audio_pkt_size);

            if (len1 < 0) {
                /* if error, we skip the frame */
                is->audio_pkt_size = 0;
                break;
            }

            is->audio_pkt_data += len1;
            is->audio_pkt_size -= len1;
            if (data_size <= 0)
                continue;

            if (dec->sample_fmt != is->audio_src_fmt) {
                if (is->reformat_ctx)
                    av_audio_convert_free(is->reformat_ctx);
                is->reformat_ctx= av_audio_convert_alloc(SAMPLE_FMT_S16, 1,
                                                         dec->sample_fmt, 1, NULL, 0);
                if (!is->reformat_ctx) {
                    fprintf(stderr, "Cannot convert %s sample format to %s sample format\n",
                        avcodec_get_sample_fmt_name(dec->sample_fmt),
                        avcodec_get_sample_fmt_name(SAMPLE_FMT_S16));
                        break;
                }
                is->audio_src_fmt= dec->sample_fmt;
            }

            if (is->reformat_ctx) {
                const void *ibuf[6]= {is->audio_buf1};
                void *obuf[6]= {is->audio_buf2};
                int istride[6]= {av_get_bits_per_sample_format(dec->sample_fmt)/8};
                int ostride[6]= {2};
                int len= data_size/istride[0];
                if (av_audio_convert(is->reformat_ctx, obuf, ostride, ibuf, istride, len)<0) {
                    printf("av_audio_convert() failed\n");
                    break;
                }
                is->audio_buf= is->audio_buf2;
                /* FIXME: existing code assume that data_size equals framesize*channels*2
                          remove this legacy cruft */
                data_size= len*2;
            }else{
                is->audio_buf= is->audio_buf1;
            }

            /* if no pts, then compute it */
            pts = is->audio_clock;
            *pts_ptr = pts;
            n = 2 * dec->channels;
            is->audio_clock += (double)data_size /
                (double)(n * dec->sample_rate);
#if defined(DEBUG_SYNC)
            {
                static double last_clock;
                printf("audio: delay=%0.3f clock=%0.3f pts=%0.3f\n",
                       is->audio_clock - last_clock,
                       is->audio_clock, pts);
                last_clock = is->audio_clock;
            }
#endif
            return data_size;
        }

        /* free the current packet */
        if (pkt->data)
            av_free_packet(pkt);

        if (is->paused || is->audioq.abort_request) {
            return -1;
        }

        /* read next packet */
        if (packet_queue_get(&is->audioq, pkt, 1) < 0)
            return -1;
        if(pkt->data == flush_pkt.data){
            avcodec_flush_buffers(dec);
            continue;
        }

        is->audio_pkt_data = pkt->data;
        is->audio_pkt_size = pkt->size;

        /* if update the audio clock with the pts */
        if (pkt->pts != AV_NOPTS_VALUE) {
            is->audio_clock = av_q2d(is->audio_st->time_base)*pkt->pts;
        }
    }
}

// SYNCH Rev. 1137
/* get the current audio output buffer size, in samples. With SDL, we
   cannot have a precise information */
static int audio_write_get_buf_size(VideoState *is)
{
    return is->audio_buf_size - is->audio_buf_index;
}

// SYNCH Rev. 14508
/* prepare a new audio buffer */
void sdl_audio_callback(void *opaque, Uint8 *stream, int len)
{
    VideoState *is = opaque;
    int audio_size, len1;
    double pts;

    audio_callback_time = av_gettime();

    while (len > 0) {
        if (is->audio_buf_index >= is->audio_buf_size) {
           audio_size = audio_decode_frame(is, &pts);
           if (audio_size < 0) {
                /* if error, just output silence */
               is->audio_buf_size = 1024;
               memset(is->audio_buf, 0, is->audio_buf_size);
           } else {
// XdTV ----------------------------------------------------------------------------
               if (audio_size>0){
                 pthread_mutex_lock(&audio_buf_mutex);
                 audiobuf=memcpy(audiobuf,is->audio_buf,audio_size); // fast_memcpy n'est pas dispo qd le thread demarre !!
                 audiobuflen=audio_size;
                 new_audiobuf_ready=1;
                 pthread_mutex_unlock(&audio_buf_mutex);
               } else {
                 pthread_mutex_lock(&audio_buf_mutex);
                 audiobuflen=0;
                 new_audiobuf_ready=0;
                 pthread_mutex_unlock(&audio_buf_mutex);
               }
// XdTV -------------------------------------------------------------------------END

// XDTV
/*
               if (is->show_audio)
                   update_sample_display(is, (int16_t *)is->audio_buf, audio_size);
*/
               audio_size = synchronize_audio(is, (int16_t *)is->audio_buf, audio_size,
                                              pts);
               is->audio_buf_size = audio_size;
               memset(is->audio_buf, 0, is->audio_buf_size); // XdTV
           }
           is->audio_buf_index = 0;
        }
        len1 = is->audio_buf_size - is->audio_buf_index;
        if (len1 > len)
            len1 = len;
        memcpy(stream, (uint8_t *)is->audio_buf + is->audio_buf_index, len1);
        len -= len1;
        stream += len1;
        is->audio_buf_index += len1;
    }
}

// SYNCH Rev. 14508
/* open a given stream. Return 0 if OK */
static int stream_component_open(VideoState *is, int stream_index)
{
    AVFormatContext *ic = is->ic;
    AVCodecContext *enc;
    AVCodec *codec;
    SDL_AudioSpec wanted_spec, spec;

    if (stream_index < 0 || stream_index >= ic->nb_streams)
        return -1;
    enc = ic->streams[stream_index]->codec;

    /* prepare audio output */
    if (enc->codec_type == CODEC_TYPE_AUDIO) {
        if (enc->channels > 0) {
            enc->request_channels = FFMIN(2, enc->channels);
        } else {
            enc->request_channels = 2;
        }
    }

    codec = avcodec_find_decoder(enc->codec_id);
    enc->debug_mv = debug_mv;
    enc->debug = debug;
    enc->workaround_bugs = workaround_bugs;
    enc->lowres = lowres;
    if(lowres) enc->flags |= CODEC_FLAG_EMU_EDGE;
    enc->idct_algo= idct;
    if(fast) enc->flags2 |= CODEC_FLAG2_FAST;
    enc->skip_frame= skip_frame;
    enc->skip_idct= skip_idct;
    enc->skip_loop_filter= skip_loop_filter;
    enc->error_recognition= error_recognition;
    enc->error_concealment= error_concealment;
    if (!codec ||
        avcodec_open(enc, codec) < 0)
        return -1;

    /* prepare audio output */
    if (enc->codec_type == CODEC_TYPE_AUDIO) {
        wanted_spec.freq = enc->sample_rate;
        wanted_spec.format = AUDIO_S16SYS;
        wanted_spec.channels = enc->channels;
        wanted_spec.silence = 0;
        wanted_spec.samples = SDL_AUDIO_BUFFER_SIZE;
        wanted_spec.callback = sdl_audio_callback;
        wanted_spec.userdata = is;
        if (SDL_OpenAudio(&wanted_spec, &spec) < 0) {
            fprintf(stderr, "SDL_OpenAudio: %s\n", SDL_GetError());
            return -1;
        }
        is->audio_hw_buf_size = spec.size;
        is->audio_src_fmt= SAMPLE_FMT_S16;
    }

    if(thread_count>1)
        avcodec_thread_init(enc, thread_count);
    enc->thread_count= thread_count;
    ic->streams[stream_index]->discard = AVDISCARD_DEFAULT;
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        is->audio_stream = stream_index;
        is->audio_st = ic->streams[stream_index];
        is->audio_buf_size = 0;
        is->audio_buf_index = 0;

        /* init averaging filter */
        is->audio_diff_avg_coef = exp(log(0.01) / AUDIO_DIFF_AVG_NB);
        is->audio_diff_avg_count = 0;
        /* since we do not have a precise anough audio fifo fullness,
           we correct audio sync only if larger than this threshold */
        is->audio_diff_threshold = 2.0 * SDL_AUDIO_BUFFER_SIZE / enc->sample_rate;

        memset(&is->audio_pkt, 0, sizeof(is->audio_pkt));
        packet_queue_init(&is->audioq);
        SDL_PauseAudio(0);
        break;
    case CODEC_TYPE_VIDEO:
        is->video_stream = stream_index;
        is->video_st = ic->streams[stream_index];

        is->frame_last_delay = 40e-3;
        is->frame_timer = (double)av_gettime() / 1000000.0;
        is->video_current_pts_time = av_gettime();

        packet_queue_init(&is->videoq);
        is->video_tid = SDL_CreateThread(video_thread, is);

        break;
    case CODEC_TYPE_SUBTITLE:
        is->subtitle_stream = stream_index;
        is->subtitle_st = ic->streams[stream_index];
        packet_queue_init(&is->subtitleq);

        //is->subtitle_tid = SDL_CreateThread(subtitle_thread, is); // XDTV
        break;
    default:
        break;
    }
    return 0;
}

// SYNCH Rev. 14508
static void stream_component_close(VideoState *is, int stream_index)
{
    AVFormatContext *ic = is->ic;
    AVCodecContext *enc;

    if (stream_index < 0 || stream_index >= ic->nb_streams)
        return;
    enc = ic->streams[stream_index]->codec;

    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        packet_queue_abort(&is->audioq);

        SDL_CloseAudio();

        packet_queue_end(&is->audioq);
        if (is->reformat_ctx)
            av_audio_convert_free(is->reformat_ctx);
        break;
    case CODEC_TYPE_VIDEO:
        packet_queue_abort(&is->videoq);

        /* note: we also signal this mutex to make sure we deblock the
           video thread in all cases */
        SDL_LockMutex(is->pictq_mutex);
        SDL_CondSignal(is->pictq_cond);
        SDL_UnlockMutex(is->pictq_mutex);

        SDL_WaitThread(is->video_tid, NULL);

        packet_queue_end(&is->videoq);
        break;
    case CODEC_TYPE_SUBTITLE:
        packet_queue_abort(&is->subtitleq);

        /* note: we also signal this mutex to make sure we deblock the
           video thread in all cases */
        SDL_LockMutex(is->subpq_mutex);
        is->subtitle_stream_changed = 1;

        SDL_CondSignal(is->subpq_cond);
        SDL_UnlockMutex(is->subpq_mutex);

        SDL_WaitThread(is->subtitle_tid, NULL);

        packet_queue_end(&is->subtitleq);
        break;
    default:
        break;
    }

    ic->streams[stream_index]->discard = AVDISCARD_ALL;
    avcodec_close(enc);
    switch(enc->codec_type) {
    case CODEC_TYPE_AUDIO:
        is->audio_st = NULL;
        is->audio_stream = -1;
        break;
    case CODEC_TYPE_VIDEO:
        is->video_st = NULL;
        is->video_stream = -1;
        break;
    case CODEC_TYPE_SUBTITLE:
        is->subtitle_st = NULL;
        is->subtitle_stream = -1;
        break;
    default:
        break;
    }
}

// SYNCH Rev. 1137
static void dump_stream_info(const AVFormatContext *s)
{
    if (s->track != 0)
        fprintf(stderr, "Track: %d\n", s->track);
    if (s->title[0] != '\0')
        fprintf(stderr, "Title: %s\n", s->title);
    if (s->author[0] != '\0')
        fprintf(stderr, "Author: %s\n", s->author);
    if (s->copyright[0] != '\0')
        fprintf(stderr, "Copyright: %s\n", s->copyright);
    if (s->comment[0] != '\0')
        fprintf(stderr, "Comment: %s\n", s->comment);
    if (s->album[0] != '\0')
        fprintf(stderr, "Album: %s\n", s->album);
    if (s->year != 0)
        fprintf(stderr, "Year: %d\n", s->year);
    if (s->genre[0] != '\0')
        fprintf(stderr, "Genre: %s\n", s->genre);
}

/* since we have only one decoding thread, we can use a global
   variable instead of a thread local variable */
// SYNCH Rev. 1137
static VideoState *global_video_state;

// SYNCH Rev. 1137
static int decode_interrupt_cb(void)
{
    return (global_video_state && global_video_state->abort_request);
}

// SYNCH Rev. 1137
/* this thread gets the stream from the disk or the network */
static int decode_thread(void *arg)
{
    VideoState *is = arg;
    AVFormatContext *ic;
    int err, i, ret, video_index, audio_index;
    AVPacket pkt1, *pkt = &pkt1;
    AVFormatParameters params, *ap = &params;

    video_index = -1;
    audio_index = -1;
    is->video_stream = -1;
    is->audio_stream = -1;
    is->subtitle_stream = -1;

    global_video_state = is;
    url_set_interrupt_cb(decode_interrupt_cb);

    memset(ap, 0, sizeof(*ap));

// XDTV
/*
    ap->width = frame_width;
    ap->height= frame_height;
    ap->time_base= (AVRational){1, 25};
    ap->pix_fmt = frame_pix_fmt;
*/

    err = av_open_input_file(&ic, is->filename, is->iformat, 0, ap);

    if (err < 0) {
        fprintf(stderr," cannot open file : %s, err=%d\n",is->filename,err); // XDTV to avoid print_error function
        ret = -1;
        goto fail;
    }
    is->ic = ic;

    if(genpts)
        ic->flags |= AVFMT_FLAG_GENPTS;

    err = av_find_stream_info(ic);
    if (err < 0) {
        fprintf(stderr, "%s: could not find codec parameters\n", is->filename);
        ret = -1;
        goto fail;
     }
    if(ic->pb)
        ic->pb->eof_reached= 0;

    if (start_time != AV_NOPTS_VALUE) {
        int64_t timestamp;

        timestamp = start_time;
        if (ic->start_time != AV_NOPTS_VALUE)
            timestamp += ic->start_time;
        ret = av_seek_frame(ic, -1, timestamp, AVSEEK_FLAG_BACKWARD);
        if (ret < 0) {
            fprintf(stderr, "%s: could not seek to position %0.3f\n",
                    is->filename, (double)timestamp / AV_TIME_BASE);
        }
    }

    for(i = 0; i < ic->nb_streams; i++) {
        AVCodecContext *enc = ic->streams[i]->codec;
        ic->streams[i]->discard = AVDISCARD_ALL;
        switch(enc->codec_type) {
        case CODEC_TYPE_AUDIO:
            if ((audio_index < 0 || wanted_audio_stream-- > 0) && !audio_disable)
                audio_index = i;
            break;
        case CODEC_TYPE_VIDEO:
            if ((video_index < 0 || wanted_video_stream-- > 0) && !video_disable)
                video_index = i;
            break;
        default:
            break;
        }
    }

// XDTV: To avoid using the show status function
//    if (show_status) {
        dump_format(ic, 0, is->filename, 0);
        dump_stream_info(ic);
//    }

    /* open the streams */
    if (audio_index >= 0) {
        stream_component_open(is, audio_index);
    }

    if (video_index >= 0) {
        stream_component_open(is, video_index);
    } else {
        if (!display_disable)
            is->show_audio = 1;
    }

    if (is->video_stream < 0 && is->audio_stream < 0) {
        fprintf(stderr, "%s: could not open codecs\n", is->filename);
        ret = -1;
        goto fail;
    }

// Begin XdTV Part ---------------------------------
    pthread_mutex_lock(&stream_open_mutex);
    stream_opened=1;
    pthread_mutex_unlock(&stream_open_mutex);
// End XdTV Part -----------------------------------

     for(;;) {
         if (is->abort_request)
             break;
        if (is->paused != is->last_paused) {
            is->last_paused = is->paused;
            if (is->paused)
                av_read_pause(ic);
            else
                av_read_play(ic);
        }

        if (is->seek_req) {
            int stream_index= -1;
            int64_t seek_target= is->seek_pos;

            if     (is->   video_stream >= 0) stream_index= is->   video_stream;
            else if(is->   audio_stream >= 0) stream_index= is->   audio_stream;
            else if(is->subtitle_stream >= 0) stream_index= is->subtitle_stream;

            if(stream_index>=0){
                seek_target= av_rescale_q(seek_target, AV_TIME_BASE_Q, ic->streams[stream_index]->time_base);
            }

            ret = av_seek_frame(is->ic, stream_index, seek_target, is->seek_flags);

            if (ret < 0) {
                fprintf(stderr, "%s: error while seeking\n", is->ic->filename);
            }else{
                if (is->audio_stream >= 0) {
                    packet_queue_flush(&is->audioq);
                    packet_queue_put(&is->audioq, &flush_pkt);
                }
                if (is->subtitle_stream >= 0) {
                    packet_queue_flush(&is->subtitleq);
                    packet_queue_put(&is->subtitleq, &flush_pkt);
                }
                if (is->video_stream >= 0) {
                    packet_queue_flush(&is->videoq);
                    packet_queue_put(&is->videoq, &flush_pkt);
                }
            }
            is->seek_req = 0;
        }

        /* if the queue are full, no need to read more */
        if (is->audioq.size > MAX_AUDIOQ_SIZE ||
            is->videoq.size > MAX_VIDEOQ_SIZE ||
            is->subtitleq.size > MAX_SUBTITLEQ_SIZE ||
            url_feof(ic->pb)) {
            /* wait 10 ms */
            SDL_Delay(10);
            continue;
        }
        ret = av_read_frame(ic, pkt);
        if (ret < 0) {
            if (url_ferror(ic->pb) == 0) {
                 SDL_Delay(100); /* wait for user event */
                continue;
            } else
                break;
        }
        if (pkt->stream_index == is->audio_stream) {
            packet_queue_put(&is->audioq, pkt);
        } else if (pkt->stream_index == is->video_stream) {
            packet_queue_put(&is->videoq, pkt);
        } else if (pkt->stream_index == is->subtitle_stream) {
            packet_queue_put(&is->subtitleq, pkt);
        } else {
            av_free_packet(pkt);
        }
    }
    /* wait until the end */
    while (!is->abort_request) {
        SDL_Delay(100);
    }

    ret = 0;
 fail:
    /* disable interrupting */
    global_video_state = NULL;

    /* close each stream */
    if (is->audio_stream >= 0)
        stream_component_close(is, is->audio_stream);
    if (is->video_stream >= 0)
        stream_component_close(is, is->video_stream);
    if (is->subtitle_stream >= 0)
        stream_component_close(is, is->subtitle_stream);
    if (is->ic) {
        av_close_input_file(is->ic);
        is->ic = NULL; /* safety */
    }
    url_set_interrupt_cb(NULL);

    if (ret != 0) {
        SDL_Event event;

        event.type = FF_QUIT_EVENT;
        event.user.data1 = is;
        SDL_PushEvent(&event);
    }
    return 0;
}

// SYNCH Rev. 1137
static VideoState *stream_open(const char *filename, AVInputFormat *iformat)
{
    VideoState *is;

// For XdTV -----------------------------------------------------------------------------------------------------------------
    pthread_mutex_init (&stream_open_mutex, NULL); // on attend que la stream soit ouverte avant de faire quoi que ce soit ...
    pthread_mutex_lock(&stream_open_mutex);
    stream_opened=0;
    pthread_mutex_unlock(&stream_open_mutex);
// For XdTV -------------------------------------------------------------------------------------------------------------END

    is = av_mallocz(sizeof(VideoState));
    if (!is)
        return NULL;
// XDTV: To avoid av_strlcpy
    strncpy((char*) is->filename, (const char*) filename, sizeof(is->filename));
    is->iformat = iformat;

    is->ytop = 0;
    is->xleft = 0;

    /* start video display */
    is->pictq_mutex = SDL_CreateMutex();
    is->pictq_cond = SDL_CreateCond();

    is->subpq_mutex = SDL_CreateMutex();
    is->subpq_cond = SDL_CreateCond();

    /* add the refresh timer to draw the picture */
    schedule_refresh(is, 40);

    is->av_sync_type = av_sync_type;
    is->parse_tid = SDL_CreateThread(decode_thread, is);
    if (!is->parse_tid) {
        av_free(is);
        return NULL;
    } 


    return is;
}

// SYNCH Rev. 1137
static void stream_close(VideoState *is)
{
    VideoPicture *vp;
    int i;
    /* XXX: use a special url_shutdown call to abort parse cleanly */
    is->abort_request = 1;
    SDL_WaitThread(is->parse_tid, NULL);

    /* free all pictures */
    for(i=0;i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
        vp = &is->pictq[i];

// XDTV: No bmp so this part is skipped
  /*      if (vp->bmp) {
            SDL_FreeYUVOverlay(vp->bmp);
            vp->bmp = NULL;
        }
*/
    }
    SDL_DestroyMutex(is->pictq_mutex);
    SDL_DestroyCond(is->pictq_cond);
    SDL_DestroyMutex(is->subpq_mutex);
    SDL_DestroyCond(is->subpq_cond);
}

// SYNCH Rev. 1137
static void stream_cycle_channel(VideoState *is, int codec_type)
{
    AVFormatContext *ic = is->ic;
    int start_index, stream_index;
    AVStream *st;

    if (codec_type == CODEC_TYPE_VIDEO)
        start_index = is->video_stream;
    else if (codec_type == CODEC_TYPE_AUDIO)
        start_index = is->audio_stream;
    else
        start_index = is->subtitle_stream;
    if (start_index < (codec_type == CODEC_TYPE_SUBTITLE ? -1 : 0))
        return;
    stream_index = start_index;
    for(;;) {
        if (++stream_index >= is->ic->nb_streams)
        {
            if (codec_type == CODEC_TYPE_SUBTITLE)
            {
                stream_index = -1;
                goto the_end;
            } else
                stream_index = 0;
        }
        if (stream_index == start_index)
            return;
        st = ic->streams[stream_index];
        if (st->codec->codec_type == codec_type) {
            /* check that parameters are OK */
            switch(codec_type) {
            case CODEC_TYPE_AUDIO:
                if (st->codec->sample_rate != 0 &&
                    st->codec->channels != 0)
                    goto the_end;
                break;
            case CODEC_TYPE_VIDEO:
            case CODEC_TYPE_SUBTITLE:
                goto the_end;
            default:
                break;
            }
        }
    }
 the_end:
    stream_component_close(is, start_index);
    stream_component_open(is, stream_index);
}

// SYNCH Rev. 1137
static void step_to_next_frame(void)
{
    if (cur_stream) {
        /* if the stream is paused unpause it, then step */
         if (cur_stream->paused)
            stream_pause(cur_stream);
    }
    step = 1;
}

// SYNCH Rev. 1137
/* handle an event sent by the GUI */
static void event_loop(void)
{
    SDL_Event event;
    double incr, pos, frac;

    for(;;) {
        SDL_WaitEvent(&event);
        switch(event.type) {
        case SDL_KEYDOWN:
            switch(event.key.keysym.sym) {
            case SDLK_ESCAPE:
            case SDLK_q:
                //do_exit(); // XDTV
                break;
            case SDLK_f:
                //toggle_full_screen(); // XDTV
                break;
            case SDLK_p:
            case SDLK_SPACE:
                //toggle_pause(); // XDTV
                break;
            case SDLK_s: //S: Step to next frame
                step_to_next_frame();
                break;
            case SDLK_a:
                if (cur_stream)
                    stream_cycle_channel(cur_stream, CODEC_TYPE_AUDIO);
                break;
            case SDLK_v:
                if (cur_stream)
                    stream_cycle_channel(cur_stream, CODEC_TYPE_VIDEO);
                break;
            case SDLK_t:
                if (cur_stream)
                    stream_cycle_channel(cur_stream, CODEC_TYPE_SUBTITLE);
                break;
            case SDLK_w:
                //toggle_audio_display(); // XDTV
                break;
            case SDLK_LEFT:
                incr = -10.0;
                goto do_seek;
            case SDLK_RIGHT:
                incr = 10.0;
                goto do_seek;
            case SDLK_UP:
                incr = 60.0;
                goto do_seek;
            case SDLK_DOWN:
                incr = -60.0;
            do_seek:
                if (cur_stream) {
                    if (seek_by_bytes) {
                        pos = url_ftell(cur_stream->ic->pb);
                        if (cur_stream->ic->bit_rate)
                            incr *= cur_stream->ic->bit_rate / 60.0;
                        else
                            incr *= 180000.0;
                        pos += incr;
                        stream_seek(cur_stream, pos, incr);
                    } else {
                        pos = get_master_clock(cur_stream);
                        pos += incr;
                        stream_seek(cur_stream, (int64_t)(pos * AV_TIME_BASE), incr);
                    }
                }
                break;
            default:
                break;
            }
            break;
        case SDL_MOUSEBUTTONDOWN:
            if (cur_stream) {
                int ns, hh, mm, ss;
                int tns, thh, tmm, tss;
                tns = cur_stream->ic->duration/1000000LL;
                thh = tns/3600;
                tmm = (tns%3600)/60;
                tss = (tns%60);
                frac = (double)event.button.x/(double)cur_stream->width;
                ns = frac*tns;
                hh = ns/3600;
                mm = (ns%3600)/60;
                ss = (ns%60);
                fprintf(stderr, "Seek to %2.0f%% (%2d:%02d:%02d) of total duration (%2d:%02d:%02d)       \n", frac*100,
                        hh, mm, ss, thh, tmm, tss);
                stream_seek(cur_stream, (int64_t)(cur_stream->ic->start_time+frac*cur_stream->ic->duration), 0);
            }
            break;
        case SDL_VIDEORESIZE:
            if (cur_stream) {
// XDTV
/*
                screen = SDL_SetVideoMode(event.resize.w, event.resize.h, 0,
                                          SDL_HWSURFACE|SDL_RESIZABLE|SDL_ASYNCBLIT|SDL_HWACCEL);
                screen_width = cur_stream->width = event.resize.w;
                screen_height= cur_stream->height= event.resize.h;
*/
                cur_stream->width = event.resize.w;
                cur_stream->height = event.resize.h;
            }
            break;
        case SDL_QUIT:
// XdTV ------------------------------------------------------------------------
	    if (debug)	
            	fprintf(stderr,"grab-dvb : event_loop thread ended here ! \n");
            return;
// XDTV ---------------------------------------------------------------------END
        case FF_QUIT_EVENT:
            //do_exit();
            break;
        case FF_ALLOC_EVENT:
            //video_open(event.user.data1); // XDTV
            alloc_picture(event.user.data1);
            break;
        case FF_REFRESH_EVENT:
            video_refresh_timer(event.user.data1);
            break;
        default:
            break;
        }
    }
    return; // XDTV
}


/**************** END FFPLAY SDL PLAYER ******************************/


/* prototypes */

int fmts_xaw_to_ffmpegcolorspace[MAX_VIDEO_FMT]={
  [VIDEO_RGB24_BE]   = PIX_FMT_RGB24,
  [VIDEO_YUV420] = PIX_FMT_YUV420P,
//  [VIDEO_YUYV]  = PIX_FMT_YUV422,
//  [VIDEO_YUV422P] = PIX_FMT_YUV422P,
0
};
  
static struct STRTAB norms[] = {
  {0,   "PAL"},
  {1,  "NTSC"},
  {2, "SECAM"},
  {-1, NULL}
};

static struct STRTAB inputs[] = {
  {0,   "Television"}, // used by default
  {1,   "Composite1"}, // needed only when the xdtvrc is read and write
  {2,   "Composite2"}, // needed only when the xdtvrc is read and write
  {3,   "Composite3"}, // needed only when the xdtvrc is read and write
  {4,   "Composite4"}, // needed only when the xdtvrc is read and write
  {5,   "S-Video"   }, // needed only when the xdtvrc is read and write
  {6,   "CSVIDEO"   }, // needed only when the xdtvrc is read and write
  {-1, NULL}
};



void* dvb_audio_main(void *arg) {

//  fprintf(stderr,"videofile->a_bits = %d\n",videofile->a_bits);
  while(1) {
	pthread_mutex_lock(&audio_buf_mutex);
	
	
 	if(new_audiobuf_ready){

    		audio_write(audiobuf,audiobuflen);
	//	write(temp_audio_file,audiobuf,audiobuflen);
		if (is_divx_in_progress()){
    			pthread_mutex_lock(&file_mutex);
      			write_audio_to_avi((short*)audiobuf, audiobuflen /(videofile->a_bits/8) );
    			pthread_mutex_unlock(&file_mutex);
    		}
		new_audiobuf_ready=0;
		pthread_mutex_unlock(&audio_buf_mutex);

  	}else {
		pthread_mutex_unlock(&audio_buf_mutex);
  		usleep(10);
	}
  	
  }
  pthread_exit(NULL);
}



void dvb_audio_start(void) {
  if(videofile==NULL) return;

  if(!audio_open(AUDIO_WO, AUDIO_S16LE,videofile->a_chans, 48000 /*videofile->a_rate */, 1 /*file_audio_fragments*/,4608 /*file_audio_sizefragment */)<0) {
     fprintf(stderr, "audio_open failed\n");
     return ;
  }
  
  mixer_set_mode(MIXER_MODE_PCM);
  if (debug)
  	fprintf(stderr,"grab-dvb : *** Create thread  : audio_thread\n");
  pthread_create (&audio_thread, NULL, dvb_audio_main, NULL);
}




static int
grab_open(struct device_t *device)
{
 
/* opening graber */
    struct stat buf;

    if (videofile==NULL)
	  videofile=malloc(sizeof(avi_t));
    videofile->a_bits=16;
    videofile->a_fmt=0x0050;
    videofile->a_chans=2;

    audio_buf_max_len=AVCODEC_MAX_AUDIO_FRAME_SIZE;
    // Allocate audio buffer
    if (NULL==(audiobuf=(char*) malloc(audio_buf_max_len))){
 	perror("malloc");
    }
   
    snprintf (FRONTEND_DEV, sizeof(FRONTEND_DEV),"/dev/dvb/adapter%i/frontend%i",devicemanager_tadapter,devicemanager_tadapter);
    snprintf (DEMUX_DEV, sizeof(DEMUX_DEV),"/dev/dvb/adapter%i/demux%i", devicemanager_tadapter , devicemanager_tadapter);
    snprintf (DVR_DEV, sizeof(DVR_DEV),"/dev/dvb/adapter%i/dvr%i", devicemanager_tadapter ,devicemanager_tadapter);

    input_filename=DVR_DEV;

    if (0 != stat (input_filename,&buf)){
    	free (videofile);
    	free (audiobuf);
	videofile=NULL;
	audiobuf=NULL;
    	return -1;
    }
    if (debug)
    	printf ("using '%s' and '%s'\n", FRONTEND_DEV, DEMUX_DEV);

    memset(&frontend_param, 0, sizeof(struct dvb_frontend_parameters));
	
    /* register all codecs, demux and protocols */
    av_register_all();
    av_register_input_format(&mpegts_demuxer);

//    pthread_mutex_lock(&grabber_mutex);
    dvb_grabber_running=1;
//    pthread_mutex_unlock(&grabber_mutex);

    return 0;
}

static int
grab_close ()
{
  
    SDL_Event event;
    event.type = SDL_QUIT ;
    SDL_PushEvent(&event);

  if (cur_stream)
	  stream_close(cur_stream);

  if (event_loop_thread_started){
  	if (debug)
	  fprintf(stderr,"closing event_loop_thread ... \n");
	  pthread_join(event_loop_thread,NULL);
	  event_loop_thread_started=0;
  	if (debug)
	  fprintf(stderr,"thread close ... \n");
  }

  free(videofile);
  videofile=NULL;

  av_free(pFrameRGB);
  free(bufferRGB);
  av_free(pFrameSRC);
  free(bufferSRC);
  free(audiobuf);
  pFrameRGB=NULL;
  bufferRGB=NULL;
  pFrameSRC=NULL;
  bufferSRC=NULL;
  audiobuf=NULL;
  audiobuflen=0;


 if (pFrameRGB_resized) av_free(pFrameRGB_resized);
 if (pFrameYUV_resized) av_free(pFrameYUV_resized);
 if (bufferRGB_resized) free(bufferRGB_resized);
 if (bufferYUV_resized) free(bufferYUV_resized);
#ifndef HAVE_SWSCALE
 if (imgResCtx) 	av_free(imgResCtx);
 imgResCtx=NULL;
#else
 if (swscale_context) 	sws_freeContext(swscale_context);
 swscale_context=NULL;
#endif

  pFrameRGB_resized=NULL;
  pFrameYUV_resized=NULL;
  bufferRGB_resized=NULL;
  bufferYUV_resized=NULL;
  width_resized=0;
  height_resized=0;
  
/***** DVB PART ****/

 if (debug)
  fprintf(stderr,"closing dvb devices\n");

  close(frontend_fd);
  close(video_fd);
  close(audio_fd);
  close(subtit_fd);
//  close(dvr_fd);
  close(vdemuxfd);
  close(ademuxfd);
  close(sdemuxfd);
  
//  pthread_mutex_lock(&grabber_mutex);
  dvb_grabber_running=0;
//  pthread_mutex_unlock(&grabber_mutex);

  return 0;
}


static int
grab_overlay (int x, int y, int width, int height, int format,
              struct OVERLAY_CLIP *oc, int count)
{
  return 0;
}

#ifndef HAVE_SWSCALE
static AVFrame * convert_img_to_requested_size (AVFrame* src_frame, int format,int width,int height , int src_width , int src_height ){

  int ret_img_conv=-1;
  if (width != width_resized || height != height_resized ) { // pFrame_resized must be re-allocated

	if (imgResCtx)		 av_free(imgResCtx);
  	if (pFrameRGB_resized) 	 av_free(pFrameRGB_resized);
	if (pFrameYUV_resized)   av_free(pFrameYUV_resized);
	if (bufferRGB_resized)	 free(bufferRGB_resized);
	if (bufferYUV_resized)   free(bufferYUV_resized);

	if (NULL==(pFrameRGB_resized=avcodec_alloc_frame())){
	       fprintf (stderr,"dvb: cannot allocate frame RGB resized\n");
	       return NULL;
	}
	if (NULL==(pFrameYUV_resized=avcodec_alloc_frame())){
	       fprintf (stderr,"dvb: cannot allocate frame YUV resized\n");
	       return NULL;
	}

  	numBytesRGB_resized = avpicture_get_size(PIX_FMT_RGB24,width,height);
  	numBytesYUV_resized = avpicture_get_size(PIX_FMT_YUV420P,width,height);

  	if (NULL == (bufferRGB_resized   = malloc(numBytesRGB_resized*sizeof(uint8_t)))){
		perror("malloc");
		return NULL;
	}
	if (NULL == (bufferYUV_resized   = malloc(numBytesYUV_resized*sizeof(uint8_t)))){
		perror("malloc");
		return NULL;
	}
	avpicture_fill((AVPicture *)pFrameRGB_resized, bufferRGB_resized, PIX_FMT_RGB24,width,height);
	avpicture_fill((AVPicture *)pFrameYUV_resized, bufferYUV_resized, PIX_FMT_YUV420P,width,height);
	width_resized=width;
	height_resized=height;
	imgResCtx = img_resample_init( width, height, src_width , src_height );

  }

  img_resample(imgResCtx ,   (AVPicture *)pFrameYUV_resized ,  (AVPicture*) (src_frame));
  ret_img_conv = img_convert((AVPicture *)pFrameRGB_resized,   PIX_FMT_RGB24,
  			     (AVPicture *)(pFrameYUV_resized), PIX_FMT_YUV420P,width,height);
  if (ret_img_conv != 0) 
  	 fprintf(stderr,"img_convert returns %d\n",ret_img_conv);	
  return pFrameRGB_resized;
}
#endif


static void * get_img (video_fmt f, int width, int height)
{
  int format ;
  format = fmts_xaw_to_ffmpegcolorspace[f];
 

 if (new_picture_ready){
	 
	pthread_mutex_lock(&new_pict_mutex);
//	fprintf(stderr,"in get_img () : new picture available\n");
	new_picture_ready=0;
	pthread_mutex_unlock(&new_pict_mutex);
		
	img++; 
#ifndef HAVE_SWSCALE
	if(pFrameSRC && (width!=videofile->width || height!=videofile->height )) {
	 	AVFrame* ret_frame_rgb;
//		fprintf(stderr,"in get_img () : conversion from %ldx%ld to %ldx%ld \n", videofile->width , videofile->height , width,height );
		if (NULL== (ret_frame_rgb=convert_img_to_requested_size( 
			    	 pFrameSRC , format, width, height , 
				 videofile->width , videofile->height ))){
			return NULL;
		}

	 	return ret_frame_rgb->data[0];

  	}else if (pFrameRGB && pFrameSRC ){
		img_convert((AVPicture *)pFrameRGB,PIX_FMT_RGB24,(AVPicture*) (pFrameSRC), PIX_FMT_YUV420P ,
			videofile->width, videofile->height);
		return pFrameRGB->data[0];
	}
#else
        if(pFrameSRC && pFrameRGB){
		// Try to see if the old context is still valid, otherwise it will realloc it cleanly
		swscale_context = sws_getCachedContext(swscale_context,
				videofile->width, videofile->height,
				PIX_FMT_YUV420P,
				width, height,
				PIX_FMT_RGB24,
				SWS_BICUBIC,
				NULL, NULL, NULL);

		sws_scale(swscale_context, 
                                ((AVPicture*) (pFrameSRC))->data, 
                                ((AVPicture*) (pFrameSRC))->linesize,
                                0, 
                                videofile->height,
				((AVPicture *)pFrameRGB)->data, 
                                ((AVPicture *)pFrameRGB)->linesize);
                return pFrameRGB->data[0];
        }
#endif
 }else {
  	return NULL;
}
  return NULL;
}

	
static int fmt_available(video_fmt f) {
    return fmts_xaw_to_ffmpegcolorspace[f]!=0;
}

/* ---------------------------------------------------------------------- */

static int
grab_tune (const struct CHANNEL* chan)
{
  int flags;
  int width,height;
  if (NULL == chan){
    fprintf(stderr , "grab-dvb : NULL chan pointer\n");
    return -1;
  }
  if (chan->channel_type != DVB_T){
    fprintf(stderr , "grab-dvb : can only tune to a DVB_T device !\n");
    return -1;
  }
  if (dvb_grabber_running)
  	grab_close();
  
  grab_open(NULL);

//  debug_channel(chan);	


//  fprintf(stderr , "grab-dvb : reinit frontend_param\n");

  frontend_param.frequency=chan->dvb_freq;
  frontend_param.inversion=chan->inversion;
  frontend_param.u.ofdm.bandwidth=chan->bandwidth;
  frontend_param.u.ofdm.code_rate_HP=chan->code_rate_high;
  frontend_param.u.ofdm.code_rate_LP=chan->code_rate_low;
  frontend_param.u.ofdm.constellation=chan->constellation;
  frontend_param.u.ofdm.transmission_mode=chan->transmission_mode;
  frontend_param.u.ofdm.guard_interval=chan->guard_interval;
  frontend_param.u.ofdm.hierarchy_information=chan->hierarchy;
  vpid=chan->vpid;
  apid=chan->apid;
  spid=chan->spid;

  if ((frontend_fd = open(FRONTEND_DEV, O_RDWR)) < 0) {
	fprintf(stderr, "failed opening %s\n", FRONTEND_DEV);
	return -1;
  }

  if (setup_frontend (frontend_fd, &frontend_param) < 0){
	fprintf(stderr, "failed setup frontend\n");
	return -1;
  }

  printf ("video pid : %04d, audio pid : %04d, subtitles pid :  %04d\n", vpid,apid,spid);

  if ((video_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
	fprintf(stderr,"failed opening %s\n", DEMUX_DEV);
	return -1;
  }

  if (set_pesfilter (video_fd, vpid, DMX_PES_VIDEO, dvr) < 0){
	fprintf(stderr,"failed set_pesfiler video\n");
	return -1;
  }

  if ((audio_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
	fprintf(stderr,"failed opening %s\n", DEMUX_DEV);
	return -1;
  }

  if (set_pesfilter (audio_fd, apid, DMX_PES_AUDIO, dvr) < 0){
	fprintf(stderr,"failed set_pesfilter audio\n");
	return -1;
  }
  if ((subtit_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
	fprintf(stderr,"failed opening %s\n", DEMUX_DEV);
	return -1;
  }

  if (set_pesfilter (subtit_fd, spid, DMX_PES_SUBTITLE, dvr) < 0){
	fprintf(stderr,"failed set_pesfilter subtitles\n");
	return -1;
  }
/*  
 *	How setting PMT_PID = 0 ???
 *
 *  if ((pmt_fd = open(DEMUX_DEV, O_RDWR)) < 0) {
	fprintf(stderr,"failed opening %s\n", DEMUX_DEV);
	return -1;
  }

  if (set_pesfilter (pmt_fd, pmt_pid, DMX_PES_PMT, dvr) < 0){
	fprintf(stderr,"failed set_pesfilter subtitles\n");
	return -1;
  }
*/



  check_frontend (frontend_fd);

//  if ((dvr_fd = open(DVR_DEV, O_RDONLY | O_NONBLOCK)) < 0) {
//	fprintf(stderr,"failed opening %s\n", DVR_DEV);
//	return -1;
//  }
  if (-1 == (vdemuxfd = add_filter(vpid, (const unsigned char*)DEMUX_DEV))){
	fprintf(stderr,"add filter failed!\n");	
	return 1;
  }
  if (-1 == (ademuxfd = add_filter(apid, (const unsigned char*)DEMUX_DEV))){
	fprintf(stderr,"add filter failed!\n");	
	return 1;
  }
  if (-1 == (sdemuxfd = add_filter(spid, (const unsigned char*)DEMUX_DEV))){
	fprintf(stderr,"add filter failed!\n");	
	return 1;
  }



  flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER;

  if (SDL_Init (flags)) {
      fprintf(stderr, "Could not initialize SDL - %s\n", SDL_GetError());
      exit(1);
  }

  cur_stream = stream_open(input_filename,&mpegts_demuxer);

  pthread_mutex_init (&new_pict_mutex, NULL);
  pthread_mutex_init (&audio_buf_mutex, NULL);
  new_picture_ready=0;
  new_audiobuf_ready=0;

  while (stream_opened !=1 ){
  	usleep(100);
  }

  width = cur_stream->video_st->codec->width;
  height= cur_stream->video_st->codec->height;

//  fprintf(stderr,"grab_tune : allocate RGB frame %dx%d\n",width,height);

  if (NULL==(pFrameRGB=avcodec_alloc_frame())){
       fprintf (stderr,"dvb: cannot allocate frame RGB\n");
       return -1;
  }
  
  if (NULL==(pFrameSRC=avcodec_alloc_frame())){
       fprintf (stderr,"dvb: cannot allocate frame SRC\n");
       return -1;
  }

  video_set_capture_size(width,height);
  videofile->width=width;
  videofile->height=height;

  // Determine required buffer size and allocate buffer
  numBytesRGB= avpicture_get_size(PIX_FMT_RGB24,width,height);
  numBytesSRC= avpicture_get_size(PIX_FMT_YUV420P,width,height);

  bufferRGB=malloc(numBytesRGB*sizeof(uint8_t));
  bufferSRC=malloc(numBytesSRC*sizeof(uint8_t));

  // Assign appropriate parts of buffer to image planes in pFrameRGB
  avpicture_fill((AVPicture *)pFrameRGB, bufferRGB, PIX_FMT_RGB24,width,height);
  avpicture_fill((AVPicture *)pFrameSRC, bufferSRC, PIX_FMT_YUV420P,width,height);

	
  if (debug)
  	fprintf(stderr,"grab-dvb : *** Create thread  : event_loop_thread\n");
  pthread_create (&event_loop_thread, NULL, (void *)event_loop, NULL);
  event_loop_thread_started=1;

  dvb_audio_start();

  
  return 0;
}

static int
grab_tuned ()
{
  return 1;
}

static int
grab_input (int input, int norm)
{
  return 0;
}

static int
grab_picture (int color, int bright, int hue, int contrast)
{
  return 0;
}

static int
grab_audio (int mute, int volume, int *mode)
{
  return 0;
}

static int is525(int norm) {
  return 0;
}

static int issecam(int norm) {
  return 0;
}

static void* get_buf(int i){
  return NULL;
}

//////////////// SPECIAL FUNCTIONS .... ///


struct GRABBER grab_dvb = {
  "dvb",
  norms, inputs,
  grab_open,
  grab_close,
  grab_overlay,
  fmt_available,
  get_img,
  grab_tune,
  grab_tuned,
  grab_input,
  grab_picture,
  grab_audio,
  is525,
  issecam,
  get_buf
};
#endif

